import {
    CalculatorType,
    ParamDefinition,
    Props,
    Observer,
    Nub,
    Session,
} from "jalhyd";

import { FormulaireElement } from "./formulaire-element";
import { Field } from "./field";
import { SelectField } from "./select-field";
import { NgParameter, ParamRadioConfig } from "./ngparam";
import { FieldsetContainer } from "./fieldset-container";
import { SelectFieldCustom } from "./select-field-custom";
import { FormulaireFixedVar } from "../definition/form-fixedvar";
import { SelectEntry } from "./select-entry";
import { FormulaireNode } from "./formulaire-node";
import { ServiceFactory } from "app/services/service-factory";

export class FieldSet extends FormulaireElement implements Observer {

    /** Nub associé */
    private _nub: Nub;

    /** fichier de configuration */
    private _jsonConfig: {};

    public get nub(): Nub {
        return this._nub;
    }

    get label(): string {
        let label = this._label;
        if (this.parent instanceof FieldsetContainer) {
            label += " n°" + String(this.parent.getFieldsetPosition(this) + 1);
        }
        return label;
    }

    public setNub(sn: Nub, update: boolean = true) {
        this._nub = sn;
        this.setParentNubForAllFields();
        if (update) {
            this.updateFields();
        }
    }

    private addField(f: Field) {
        if (! f) {
            throw new Error("FieldSet.addField() : argument incorrect (undefined)");
        }
        this.kids.push(f);
    }

    /**
     * backup NgParameters kids
     * @returns a list of object representation with sybol, value, modification flag
     */
    public backupParameters(): any[] {
        const res: any[] = [];
        // for (const p of this.allFormElements) {
        for (const p of this._kids) {
            if (p instanceof NgParameter) {
                res.push(p.objectRepresentation());
            }
        }
        return res;
    }

    /**
     * restore NgParameters kids from backup list
     * @param backup list of NgParameter object representation
     * @see backupParameters
     */
    public restoreParameters(backup: any[]) {
        // for (const p of this.allFormElements) {
        for (const p of this._kids) {
            if (p instanceof NgParameter) {
                for (const bp of backup) {
                    if (p.symbol === bp.prmDef.symbol) {
                        p.loadObjectRepresentation(bp, false);
                        break;
                    }
                }
            }
        }
    }

    public get hasInputs(): boolean {
        for (const f of this.kids) {
            if (f instanceof NgParameter) {
                return true;
            }
        }
        return false;
    }

    public getInput(i: number): NgParameter {
        let n = 0;
        for (const f of this.kids) {
            if (f instanceof NgParameter) {
                if (n === i) {
                    return f;
                }
                n++;
            }
        }
    }

    private parse_select(json: {}): SelectField {
        const res: SelectField = new SelectField(this);
        res.parseConfig(json);
        res.addObserver(this);
        return res;
    }

    private parse_select_custom(json: {}): SelectField {
        const source = json["source"];
        if (source === undefined || source === "") {
            throw new Error(`Fieldset.parse_select_custom(): "source" must not be empty`);
        }
        const res: SelectField = new SelectFieldCustom(this);
        res.parseConfig(json);
        res.addObserver(this);
        return res;
    }

    public get properties(): Props {
        return this.nub.properties;
    }

    private get sectionProperties(): Props {
        const section = this.nub.getChildren()[0];
        if (section) {
            return section.properties;
        } else {
            return new Props();
        }
    }

    /**
     * @param inSection if true, will look for the required property in the Nub's section (children[0])
     */
    private getPropValue(key: string, inSection: boolean = false): any {
        if (inSection) {
            return this.sectionProperties.getPropValue(key);
        } else {
            return this.properties.getPropValue(key);
        }
    }

    public setPropValue(key: string, val: any): boolean {
        return this.properties.setPropValue(key, val, this);
    }

    private getNubParamFromSymbol(symbol: string): ParamDefinition {
        return this._nub.getParameter(symbol);
    }

    public get jsonConfig(): {} {
        return this._jsonConfig;
    }

    private setParentNubForAllFields() {
    }

    /**
     * crée un input
     * @param json definition de l'input, extrait du fichier de conf du module de calcul
     * @param node_type_filter filtre sur le type de noeud (input créé si undefined ou égal)
     * @param default_radio_config config du radio fixé/à varier/à calculer
     */
    private parse_input(json: {}): NgParameter {
        const input_id: string = json["id"];
        let res: NgParameter;
        let nubParam: ParamDefinition;
        try {
            nubParam = this.getNubParamFromSymbol(input_id);
            if (nubParam.visible) {
                res = new NgParameter(nubParam, this);
                res.parseConfig(json);
            }
        } catch (e) {
            // parameter does not exist in Nub, it's normal, simply do not show it (silent fail)
        }
        return res;
    }

    private parseFields() {
        // clear everything so that parseFields() is idempotent
        this.clearFields();
        // parse children fields from config
        const fields = this._jsonConfig["fields"];
        for (const field_index in fields) {
            let field = fields[field_index];
            // detect short input syntax
            if (typeof field === "string") {
                field = {
                    "id": field
                };
            }
            const type = field["type"] || "input"; // "input" by default
            let param: Field;

            switch (type) {
                case "input":
                    param = this.parse_input(field);
                    // potentiellement undefined car certaines définitions de FieldSet comportent des paramètres
                    // qui ne sont pas tous affichés en même temps (cf. ouvrages //)
                    if (param) {
                        this.addField(param);
                    }
                    break;

                case "select":
                    param = this.parse_select(field);
                    this.addField(param);
                    break;

                case "select_custom":
                    param = this.parse_select_custom(field);
                    this.addField(param);
                    break;

            }
        }
    }

    private clearFields() {
        for (const n of this.kids) {
            n.removeObserver(this);
        }

        this.clearKids();
    }

    /**
     * Reloads the model values and properties, and reloads localisation strings
     */
    public updateFields() {
        this.parseFields();
        this.updateLocalisation();

        // for all select fields known by the form, set selected value
        // from associated property
        if (this.parentForm instanceof FormulaireFixedVar) {
            const selectIds = this.parentForm.selectids;
            for (const sId of selectIds) {
                this.setSelectValueFromProperty(sId, (this._confId === "fs_section"));
            }
        }
    }

    /**
     * Reflects a property value in the interface, through the value of a <select> field, if this select exists
     * @param inSection if true, will look for the required property in the Nub's section (children[0])
     */
    private setSelectValueFromProperty(selectId: string, inSection: boolean = false) {
        const selectField: SelectField = this.getFormulaireNodeById(selectId) as SelectField;
        if (selectField) {
            let propVal: any = this.getPropValue(selectField.associatedProperty, inSection);
            if (propVal === undefined) {
                propVal = ""; // clodo bullet-proof loading
            }
            const selectElement = selectField.getSelectedEntryFromValue(propVal);
            try {
                selectField.setValue(selectElement);
            } catch (e) {
                // silent fail, to avoid errors being thrown in Structure Nubs when
                // trying to apply same LoiDebit on new StructureType
                // console.error(`setSelectValueFromProperty: cannot set value ${propVal} on <select> ${selectId}`);
            }
        }
    }

    /**
     * Sets Nub default properties from config file, unless properties values are already set
     * (when deserialising an existing Nub)
     * @TODO default values should be held by model, not interface !
     */
    public parseConfig(json: {}, data?: {}) {
        this._jsonConfig = json;

        this._confId = json["id"];
        this._helpLink = json["help"];

        const ct: string = json["calcType"];
        const currentCt = this.properties.getPropValue("calcType");
        const calc_type: CalculatorType = currentCt ? currentCt : (ct ? CalculatorType[ct] : this.parentForm.calculatorType);
        this.setPropValue("calcType", calc_type);

        // parse fields once, so that SelectField elements are present
        // when setting default properties below
        this.parseFields();

        // for all select fields known by the form, apply default value
        // to associated property, usually from associated enum
        if (this.parentForm instanceof FormulaireFixedVar) {
            const selectIds = this.parentForm.selectids;
            for (const sId of selectIds) {
                // find select element in parent form
                const fe = this.getFormulaireNodeById(sId);
                if (fe) {
                    const prop = (fe as SelectField).associatedProperty;
                    const defaultValue = (fe as SelectField).defaultValue;
                    // Sets Nub default property, unless this property is already set
                    const currentValue = this.properties.getPropValue(prop);
                    if (defaultValue !== undefined && currentValue === undefined) {
                        let formalValue = defaultValue;
                        // !! property names must be unique throughout JaLHyd !!
                        const enumClass = Session.enumFromProperty[prop];
                        if (enumClass) {
                            formalValue = enumClass[defaultValue];
                        }
                        this.setPropValue(prop, formalValue);
                    }
                }
            }
        }

        this.updateFields();
    }

    public getNodeParameter(symbol: string): NgParameter {
        for (const p of this.kids) {
            if (p instanceof NgParameter) {
                if (p.isDisplayed && p.symbol === symbol) {
                    return p;
                }
            }
        }
    }

    public getNodeParameterValue(symbol: string): number {
        const p = this.getNodeParameter(symbol);
        if (! p) {
            throw new Error(`FieldSet.getNodeParameterValue() : pas de paramètre ${symbol} trouvé`);
        }

        switch (p.radioState) {
            case ParamRadioConfig.FIX:
                return p.getValue();

            case ParamRadioConfig.VAR:
            case ParamRadioConfig.CAL:
                return undefined;
        }
    }

    /**
     * retourne la valeur actuellement sélectionnée d'un SelectField (qui doit être affiché)
     * @param selectFieldId id du SelectField
     * @returns valeur courante du select sans le préfixe
     */
    public getSelectedValue(selectFieldId: string): string | string[] {
        for (const p of this.kids) {
            if (p instanceof SelectField && p.isDisplayed && p.id === selectFieldId) {
                if (! p.multiple) {
                    const value: string = (p.getValue() as SelectEntry).value;
                    return FormulaireElement.removePrefix(value, selectFieldId + "_");
                } else {
                    const values: string[] = (p.getValue() as SelectEntry[]).map((se) => {
                        return FormulaireElement.removePrefix(se.value, selectFieldId + "_");
                    });
                    return values;
                }
            }
        }
    }

    // interface Observer

    update(sender: any, data: any) {
        if (data.action) {
            switch (data.action) {
                case "select":
                    const senderId: string = sender.id.replace(/\d+$/, "");
                    if (senderId === "select_section") {
                        // sections paramétrées, courbes de remous, régimes uniformes
                        // "nodeType" is a property of the section child, not of the parent
                        const oldNodeType = this.nub.getChildren()[0].properties.getPropValue("nodeType");
                        if (oldNodeType !== data.value.value) { // avoid infinite loops
                            // manually notify parent so that it replaces the child Nub @WARNING clodo trick
                            this.parentForm.update(this, {
                                action: "propertyChange",
                                name: "nodeType",
                                value: data.value.value
                            });
                        }
                    } else {
                        if (this.parentForm instanceof FormulaireFixedVar) {
                            // for all select fields known by the form, apply received value
                            // to associated property
                            const selectIds = this.parentForm.selectids;
                            for (const sId of selectIds) {
                                if (senderId === sId) {
                                    // find select element in parent form
                                    const fe = this.parentForm.getFieldById(sId);
                                    if (fe && data.value !== undefined) {
                                        const prop = (fe as SelectField).associatedProperty;
                                        // for multiple select
                                        if (Array.isArray(data.value)) {
                                            this.setPropValue(prop, data.value.map((v: any) => v.value));
                                        } else {
                                            this.setPropValue(prop, data.value.value);
                                        }
                                    }
                                }
                            }
                        }
                    }

                    break; // switch (data.action)
            }
        }
    }
}
